(require-library "synrule.ss")
(require-library "functio.ss")


;bugs:
;- timeout ignored

;--- Consts etc.
(define icfp4-timeout 0)
(define icfp4-timeover 0)

(define icfp4-tags-textstyle
  (list "B" "EM" "I" "PL" "S" "TT" "U"))
(define icfp4-tags-size
  (list "0" "1" "2" "3" "4" "5" "6" "7" "8" "9"))
(define icfp4-tags-colour
  (list "r" "g" "b" "c" "m" "y" "k" "w"))
(define icfp4-tags
  (append icfp4-tags-textstyle icfp4-tags-size icfp4-tags-colour))

(define icfp4-tags-sideeffects
  ; Flags: on, off, x, tgl
  ;                B     EM    I     S     TT    U
  (list (list "B"  'on   #f    #f    #f    #f    )
        (list "I"  #f    #f    'on   #f    #f    )
        (list "EM" #f    'tgl  #f    'off    #f  )
        (list "S"  #f    'off  #f    'on   #f    )
        (list "TT" #f    #f    #f    #f    'on   )))
;  (list (list "B"  'on   #f    #f    #f    #f    #f   )
;        (list "EM" #f    'tgl  #f    'x    #f    #f   )
;        (list "I"  #f    #f    'on   #f    #f    #f   )
;        (list "S"  #f    'off  #f    'on   #f    #f   )
;        (list "TT" #f    #f    #f    #f    'on   #f   )
;        (list "U"  #f    #f    #f    #f    #f    'on  )
;        (list "PL" 'off  'off  'off  'off  'off  'off )))
(define (icfp4-tags-sideeffect-b entry) (list-ref entry 1))
(define (icfp4-tags-sideeffect-em entry) (list-ref entry 2))
(define (icfp4-tags-sideeffect-i entry) (list-ref entry 3))
(define (icfp4-tags-sideeffect-s entry) (list-ref entry 4))
(define (icfp4-tags-sideeffect-tt entry) (list-ref entry 5))
(define (icfp4-tags-sideeffect-u entry) (list-ref entry 6))

(define icfp4-whitespace ;; CR | LF | TAB | SPC
  (list #\tab
        #\newline
        #\return
        #\space))

(define icfp4-tags-start #\<)
(define icfp4-tags-end #\>)
(define icfp4-tags-close #\/)

(define icfp4-tags-start$ (string icfp4-tags-start))
(define icfp4-tags-end$   (string icfp4-tags-end))
(define icfp4-tags-close$ (string icfp4-tags-close))



;--- Props
(define-syntax icfp4-prop-define
  (syntax-rules ()
                ((icfp4-prop-define name)
                 (define name null))
                ((icfp4-prop-define name init)
                 (define name (list init)))))
(define-syntax icfp4-prop-reset!
  (syntax-rules ()
                ((icfp4-prop-reset! name)
                 (set! name null))
                ((icfp4-prop-reset! name init)
                 (set! name (list init)))))
(define-syntax icfp4-prop-push!
  (syntax-rules ()
                ((icfp4-prop-push name elem)
                 (set! name (cons elem name)))))
(define-syntax icfp4-prop-pop!
  (syntax-rules ()
                ((icfp4-prop-push name)
                 (if (null? name)
                     null
                     (let ((rv (car name)))
                       (if (> (length name) 1)
                           (set! name (cdr name)))
                       rv)))))
(define (icfp4-prop-get name)
  (cond ((null? name) 
         null)
        ((or (list? name) (pair? name))
         (car name))
        (else 
         name)))
(define (icfp4-prop-ref name pos)
  (let* ((len (length name))
         (pos (cond ((or (>= pos len) (equal? pos 'end))
                     (- len 1))
                    ((< pos 0)
                     0)
                    ((integer? pos) pos)
                    (else 0))))
    (list-ref name pos)))
(define (icfp4-prop-ref-v1 name pos . default)
  (if (and (>= pos 0)
           (<  pos (length name)))
      (list-ref name pos)
      (if (null? default)
          null
          (car default))))

(icfp4-prop-define icfp4-prop-b #f)
(icfp4-prop-define icfp4-prop-em #f)
(icfp4-prop-define icfp4-prop-i #f)
(icfp4-prop-define icfp4-prop-s #f)
(icfp4-prop-define icfp4-prop-tt #f)
(icfp4-prop-define icfp4-prop-u 0)
(icfp4-prop-define icfp4-prop-size #f)
(icfp4-prop-define icfp4-prop-colour #f) ;"w"

(define (icfp4-prop-reset)
  (icfp4-prop-reset! icfp4-prop-b #f)
  (icfp4-prop-reset! icfp4-prop-em #f)
  (icfp4-prop-reset! icfp4-prop-i #f)
  (icfp4-prop-reset! icfp4-prop-s #f)
  (icfp4-prop-reset! icfp4-prop-tt #f)
  (icfp4-prop-reset! icfp4-prop-u 0)
  (icfp4-prop-reset! icfp4-prop-size #f)
  (icfp4-prop-reset! icfp4-prop-colour #f))

(define (icfp4-state-snapshot)
  (list (icfp4-prop-get icfp4-prop-b)
        (icfp4-prop-get icfp4-prop-em)
        (icfp4-prop-get icfp4-prop-i)
        (icfp4-prop-get icfp4-prop-s)
        (icfp4-prop-get icfp4-prop-tt)
        (icfp4-prop-get icfp4-prop-u)
        (icfp4-prop-get icfp4-prop-size)
        (icfp4-prop-get icfp4-prop-colour)))

(define (icfp4-state-b snapshot)      (list-ref snapshot 0))
(define (icfp4-state-em snapshot)     (list-ref snapshot 1))
(define (icfp4-state-i snapshot)      (list-ref snapshot 2))
(define (icfp4-state-s snapshot)      (list-ref snapshot 3))
(define (icfp4-state-tt snapshot)     (list-ref snapshot 4))
(define (icfp4-state-u snapshot)      (list-ref snapshot 5))
(define (icfp4-state-size snapshot)   (list-ref snapshot 6))
(define (icfp4-state-colour snapshot) (list-ref snapshot 7))

(define (icfp4-state-diffpattern count snapshot1 snapshot2)
  (if (or (null? snapshot1) (= count 0))
      null
      (let* ((hd1 (car snapshot1))
             (hd2 (car snapshot2))
             (change? (case hd2
                        (('on #t)
                         (if (or (equal? hd1 'off)
                                 (equal? hd1 #f))
                             'on
                             #f))
                        (('off #f)
                         (if (or (equal? hd1 'on)
                                 (equal? hd1 #t))
                             'off
                             #f)))))
        (cons change?
              (icfp4-state-diffpattern (- count 1)
                                       (cdr snapshot1)
                                       (cdr snapshot2))))))

(define icfp4-in-list null)
(define icfp4-out-struct (list (icfp4-state-snapshot)))
(icfp4-prop-define icfp4-open-tags #f)



;--- Tools
(define (icfp4-error . txt)
  #f)
;  (display txt)
;  (newline))

(define (icfp4-parse-cmdline)
  (let ((timeout (if (>= (vector-length argv) 1)
                     (string->number (vector-ref argv 0))
                     null)))
    (if (number? timeout)
        (begin
          (set! icfp4-timeout timeout)
          (set! icfp4-timeover (+ (current-seconds) timeout)))
        (icfp4-error "Missing timeout argument"))))

(define (icfp4-expired?)
  (> (current-seconds) icfp4-timeover))

(define (icfp4-timepassed)
  (- icfp4-timeover (current-seconds)))

(define (icfp4-whitespace? ch)
  (list? (member ch icfp4-whitespace)))

(define (icfp4-count-member items lst)
  (let ((count 0))
    (for-each (lambda (x) (if (member x lst)
                              (set! count (+ count 1))))
              (if (list? items) items (list items)))
    count))

(define (icfp4-replace-nth lst pos repl)
  (cond ((null? lst)
         null)
        ((equal? pos 'end)
         (icfp4-replace-nth lst (- (length lst) 1) repl))
        ((or (equal? pos 'first) (= pos 0))
         (cons repl (cdr lst)))
        ((or (> (+ pos 1) (length lst)) (< pos 0))
         lst)
        (else
         (cons (car lst)
               (icfp4-replace-nth (cdr lst) (- pos 1) repl)))))

(define (icfp4-listwrap e)
  (if (list? e)
      e
      (list e)))
  


;--- Read-in
(define icfp4-read-text null)
(define icfp4-read-counter 0)
(define icfp4-read-max 0)
(define (icfp4-read-length) (length icfp4-read-text))
(define (icfp4-read-get-text)
  (if (null? icfp4-read-text)
      ""
      (string->list (reverse icfp4-read-text))))

(define (icfp4-read-init)
  #f)

(define (icfp4-read-char)
  (read-char))

(define (icfp4-read-char-disabled)
  (let ((rv (read-char)))
    (append! icfp4-read-text (list rv))
    rv))

(define (icfp4-peek-char)
  (peek-char))

(define (icfp4-eof-object? obj)
  (eof-object? obj))



;--- Input
(define (icfp4-readall-tokenized)
  (let ((alt #f)
        (have-to-read-tag? #f)
        (eof? #f))
    (letrec ((icfp4-readall-tokenized-this
              (lambda ()
                (let ((ch (icfp4-read-char)))
                  (cond ((eof-object? ch) (set! eof? #t) null)
                        ((char=? ch icfp4-tags-start) 
                         (set! have-to-read-tag? #t) null)
                        ((icfp4-whitespace? ch) 
                         (set! alt (icfp4-check-whitespace ch)) null)
                        (else
                         (cons ch (icfp4-readall-tokenized-this))))))))
      (let ((this (icfp4-readall-tokenized-this)))
        (if eof?
            null
            (let ((thisblock
                   (if (null? this) null
                       (list (icfp4-block-def 'text (list->string this)))))
                  (altblock (if alt (list alt) null)))
              (if have-to-read-tag?
                  (icfp4-in-nexttag))
              (append thisblock altblock (icfp4-readall-tokenized))))))))

(define (icfp4-check-whitespace ch)
  ; if it is a whitespace character, compute the current space context, 
  ; which is the current context modified in the following way: 
  ; the s, em, i, and b flags are unset, and if u is 0 then colour is set to w.
  ; If the tt flag is true in the current context, then output the input 
  ; character decorated by the current space context; else if the previous 
  ; character was a SPC decorated with the same space context then do nothing; 
  ; otherwise, output a SPC decorated with the current space context.
  
  ; There is another interaction between whitespace and flags: the EM, I, B, 
  ; and S attributes are irrelevant for whitespace; moreover, color is 
  ; irrelevant where U = 0.
  (letrec ((check-whitespace
            (lambda (ch)
              (if (icfp4-prop-get icfp4-prop-tt)
                  (let ((nch (icfp4-peek-char)))
                    (if (icfp4-whitespace? nch)
                        (begin (icfp4-read-char)
                               (cons ch (check-whitespace nch)))
                        (list ch)))
                  (let loop ()
                    (if (icfp4-whitespace? (icfp4-peek-char))
                        (begin
                          (icfp4-read-char)
                          (loop))
                        (list #\space)))))))
    (let ((setsc? (and (not (icfp4-prop-get icfp4-prop-s))
                       (not (icfp4-prop-get icfp4-prop-em))
                       (not (icfp4-prop-get icfp4-prop-i))
                       (not (icfp4-prop-get icfp4-prop-b))
                       (or (not (icfp4-prop-get icfp4-prop-u))
                           (= (icfp4-prop-get icfp4-prop-u) 0)))))
      ;      (if setsc? (icfp4-change-prop-colour 'opentag "w"))
      (let ((rv (icfp4-block-def 'whitespace
                                 (list->string (check-whitespace ch)))))
        ;        (if setsc? (icfp4-change-prop-colour 'closetag "w"))
        rv))))

(define (icfp4-in-nexttag)
  (let loop ((opentag? #t)
             (tag null)
             (ch (icfp4-read-char)))
    (cond ((char=? ch icfp4-tags-close)
           (loop #f tag (icfp4-read-char)))
          ((not (char=? ch icfp4-tags-end))
           (loop opentag?
                 (append tag (list ch))
                 (icfp4-read-char)))
          (else
           (let ((thistag (list->string tag)))
             (if opentag?
                 (icfp4-change-prop 'opentag thistag)
                 (icfp4-change-prop 'closetag thistag)))))))

(define (icfp4-block-def type text)
  (list type (icfp4-state-snapshot) text))
(define (icfp4-block-get-type block)
  (car block))
(define (icfp4-block-get-prop block)
  (cadr block))
(define (icfp4-block-get-text block)
  (cddr block))

(define (icfp4-block-text? block)
  (equal? (icfp4-block-get-type block) 'text))
(define (icfp4-block-opentag? block)
  (equal? (icfp4-block-get-type block) 'opentag))
(define (icfp4-block-closetag? block)
  (equal? (icfp4-block-get-type block) 'closetag))
(define (icfp4-block-whitespace? block)
  (equal? (icfp4-block-get-type block) 'whitespace))

(define (icfp4-opentag? type)
  (or (equal? type 'opentag)
      (equal? type #t)))
(define (icfp4-closetag? type)
  (equal? type 'closetag))

(define-syntax icfp4-settag!
  (syntax-rules ()
                ((icfp4-settag! name type)
                 (if (icfp4-opentag? type)
                     (icfp4-prop-push! name #t)
                     (icfp4-prop-pop! name)))
                ((icfp4-settag! name type val)
                 (if (icfp4-opentag? type)
                     (icfp4-prop-push! name val)
                     (icfp4-prop-pop! name)))))

(define-syntax icfp4-toggletag!
  (syntax-rules ()
                ((icfp4-settag! name type)
                 (if (icfp4-opentag? type)
                     (icfp4-prop-push! name (not (icfp4-prop-get name)))
                     (icfp4-prop-pop! name)))))

(define (icfp4-change-prop type text)
  (cond ((not text)
         #f)
        ((member text icfp4-tags-size)
         (icfp4-change-prop-size type text))
        ((member text icfp4-tags-colour)
         (icfp4-change-prop-colour type text))
        ((string=? text "B")
         (icfp4-change-prop-b type))
        ((string=? text "EM")
         (icfp4-change-prop-em type))
        ((string=? text "I")
         (icfp4-change-prop-i type))
        ((string=? text "PL")
         (icfp4-change-prop-pl type))
        ((string=? text "S")
         (icfp4-change-prop-s type))
        ((string=? text "TT")
         (icfp4-change-prop-tt type))
        ((string=? text "U")
         (icfp4-change-prop-u type))
        (else
         (icfp4-error "Unknown tag" text type))))

(define (icfp4-change-prop-colour type text)
  (icfp4-settag! icfp4-prop-colour type text))

(define (icfp4-change-prop-size type text)
  (icfp4-settag! icfp4-prop-size type (string->number text)))

(define (icfp4-change-prop-u type)
  ;if u is less than 3, increment it; otherwise do nothing 
  (let ((u (icfp4-prop-get icfp4-prop-u)))
    (if (< u 3)
        (icfp4-settag! icfp4-prop-u type
                       (if (number? u)
                           (+ u 1)
                           1)))))

(define (icfp4-change-prop-b type)
  ;set the b flag
  (icfp4-settag! icfp4-prop-b type))

(define (icfp4-change-prop-em type)
  ;invert the em flag if the s flag is not set; otherwise do nothing 
  (if (not (icfp4-prop-get icfp4-prop-s))
      (icfp4-toggletag! icfp4-prop-em type)))

(define (icfp4-change-prop-i type)
  ;set the i flag 
  (icfp4-settag! icfp4-prop-i type))

(define (icfp4-change-prop-pl type)
  ;unset the u, s, em, i, b, and tt flags 
  (icfp4-settag! icfp4-prop-u type #f)
  (icfp4-settag! icfp4-prop-s type #f)
  (icfp4-settag! icfp4-prop-em type #f)
  (icfp4-settag! icfp4-prop-i type #f)
  (icfp4-settag! icfp4-prop-b type #f)
  (icfp4-settag! icfp4-prop-tt type #f))

(define (icfp4-change-prop-s type)
  ;set the s flag and unset the em flag 
  (icfp4-settag! icfp4-prop-s type)
  (icfp4-settag! icfp4-prop-em type #f))

(define (icfp4-change-prop-tt type)
  ;set the tt flag 
  (icfp4-settag! icfp4-prop-tt type))



;--- Output
(define (icfp4-output input)
  (let ((timer (icfp4-timepassed)))
    (icfp4-prop-reset)
    ; (display "input:")(newline)(display input)(newline)
    (let* ((out (icfp4-optimize-whitespaces-phase2 input #f))
           ; (out (begin (display out)(newline) out))
           (out (icfp4-output-list out))
           ; (out (begin (display out)(newline) out))
           (out (icfp4-optimize-redundances null
                                            (icfp4-list-ref 0 out)
                                            (icfp4-list-ref 1 out)
                                            (icfp4-nth-dr   1 out)))
           ; (out (begin (display out)(newline) out))
           (out (icfp4-optimize-pl out))
           ; (out (begin (display out)(newline) out))
           (out (icfp4-optimize-colour out))
           ; (out (begin (display out)(newline) out))
           (out (icfp4-optimize-size out))
           ; (out (begin (display out)(newline) out))
           (out (icfp4-optimize-checkorder out)))
      ; (display "outlst:")(display out)(newline)
      ; (display "output:")(newline)
      (icfp4-display out))))

(define (icfp4-display out)
  (if (not (null? out))
      (let* ((block (car out))
             (pre   (icfp4-outblock-pre block))
             (text  (icfp4-outblock-text block))
             (post  (icfp4-outblock-post block))
             (post  (if (pair? post)
                        (reverse post)
                        post))
             (tagger (lambda (type x)
                       (if (null? x)
                           null
                           (icfp4-build-tag type x))))
             (displayer (lambda (x)
                          (cond ((null? x) null)
                                ((pair? x) (map display x))
                                (else      (display x))))))
        (displayer (tagger 'opentag pre))
        ;        (display "[")(display block)(display "]")
        (if (pair? text)
            (if (pair? (car text))
                (map icfp4-display text)
                (displayer text)))
        (displayer (tagger 'closetag post))
        (icfp4-display (cdr out)))))

(define (icfp4-outblock-new pre text post) 
  (list (icfp4-listwrap pre) text (icfp4-listwrap post)))
(define (icfp4-outblock-pre block)  (list-ref block 0))
(define (icfp4-outblock-text block) (list-ref block 1))
(define (icfp4-outblock-post block) (list-ref block 2))
(define (icfp4-outblock-patch block pre text post)
  (let ((pre   (or pre  (icfp4-outblock-pre block)))
        (text  (or text (icfp4-outblock-text block)))
        (post  (or post (icfp4-outblock-post block))))
    (icfp4-outblock-new (icfp4-listwrap pre) text (icfp4-listwrap post))))

(define (icfp4-list-ref nth lst)
  (cond ((or (null? lst) (< nth 0) (>= nth (length lst)))
         null)
        (else
         (list-ref lst nth))))

(define (icfp4-nth-dr nth lst)
  (cond ((null? lst) null)
        ((<= nth 0)  (cdr lst))
        (else        (icfp4-nth-dr (- nth 1) (cdr lst)))))

(define (icfp4-output-list lst)
  (if (null? lst)
      null
      (let* ((block      (car lst))
             (rest       (cdr lst))
             (type       (icfp4-block-get-type block))
             (prop       (icfp4-block-get-prop block))
             (text       (icfp4-block-get-text block))
             (state      (icfp4-state-snapshot))
             (tags       (icfp4-which-tags state prop)))
        (cons (icfp4-outblock-new tags text tags)
              (icfp4-output-list rest)))))

(define (icfp4-which-tags state prop)
  (let* ((col-old (icfp4-state-colour state))
         (col-new (icfp4-state-colour prop))
         (col-changed? (not (equal? col-old col-new)))
         (col-tags (if col-changed?
                       (icfp4-which-col-tags col-old col-new)
                       null))
         (size-old (icfp4-state-size state))
         (size-new (icfp4-state-size prop))
         (size-changed? (not (equal? size-old size-new)))
         (size-tags (if size-changed?
                        (icfp4-which-size-tags size-old size-new)
                        null))
         (ul-old (icfp4-state-u state))
         (ul-new (icfp4-state-u prop))
         (ul-changed? (not (equal? ul-old ul-new)))
         (ul-tags (if ul-changed?
                      (icfp4-which-ul-tags ul-old ul-new)
                      null))
         (style-diff (build-list 5 (lambda (x) (list-ref prop x))))
         (style-state (build-list 5 (lambda (x) (list-ref state x))))
         (style-changed? (icfp4-which-style-tags-changed
                          style-state style-diff))
         (style-tags (if style-changed?
                         (icfp4-which-style-tags style-diff style-state)
                         null)))
    (append col-tags size-tags ul-tags style-tags)))

(define (icfp4-which-col-tags col-old col-new)
  (list col-new))

(define (icfp4-which-size-tags size-old size-new)
  (list size-new))

(define (icfp4-which-ul-tags ul-old ul-new)
  (build-list (abs (- (if (number? ul-old) ul-old 0)
                      (if (number? ul-new) ul-new 0)))
              (lambda (x) "U")))

(define (icfp4-which-style-tags-changed state prop)
  (ormap (lambda (a b) (not (equal? a b))) state prop))

(define (icfp4-which-style-tags diff state)
  (if (icfp4-which-style-tags-changed diff state)
      (let ((tag-and (icfp4-find-tag andmap diff icfp4-tags-sideeffects)))
        (if tag-and
            (list tag-and)
            (let ((tag-or (icfp4-find-tag ormap diff icfp4-tags-sideeffects)))
              (if tag-or
                  (let ((diff1 (icfp4-minus-style-tag diff tag-or)))
                    ; (display "[")(display tag-or)(display ":")
                    ; (display diff)(display "->")
                    ; (display diff1)(display "]")(newline)
                    (cons tag-or (icfp4-which-style-tags diff1 state)))
                  ;;; oops
                  null))))
      null))

(define (icfp4-find-tag fun style-diff lst)
  (if (null? lst)
      #f
      (let* ((entry (car lst))
             (id    (car entry))
             (mask  (cdr entry))
             (ok    (icfp4-compare-mask fun style-diff mask)))
        (if ok
            id
            (icfp4-find-tag fun style-diff (cdr lst))))))

(define (icfp4-compare-mask fun diff mask)
  (fun icfp4-match-mask diff mask))

;(define (icfp4-match-mask diff mask)
;  (case mask
;    ((on) 
;     (or (equal? diff 'on)
;         (and (boolean? diff) diff)))
;    ((off x #f)
;     (or (equal? diff 'off)
;         (not diff)))
;    ((tgl)
;     (or (equal? diff 'on)
;         (equal? diff 'off)
;         (and (boolean? diff) diff)))
;    (else (or (not (equal? diff 'on))
;              (not (equal? diff 'off))
;              (equal? diff mask)))))

;(define (icfp4-match-mask diff mask)
;  (if (number? diff)
;      (or (equal? mask #f)
;          (number? diff))
;      (case diff
;        ((on #t) 
;         (or (equal? mask 'on) (equal? mask #t) (equal? mask 'tgl)))
;        ((off #f)
;;         (or (equal? mask 'off) (equal? mask #f)))
;         (equal? mask 'off))
;        (else (equal? diff mask)))))

(define (icfp4-match-mask diff mask)
  (if (number? diff)
      (or (equal? mask #f)
          (number? diff))
      (if diff
          (or (equal? mask 'on) (equal? mask #t) (equal? mask 'tgl))
          #f)))

(define (icfp4-minus-style-tag diff id)
  (let* ((mask (assoc id icfp4-tags-sideeffects))
         (mask (if (pair? mask)
                   (cdr mask)
                   #f))
         (minus (lambda (diff mask)
                  (case mask
                    ((on tgl) (if (or (equal? diff 'on) 
                                  (equal? diff #t))
                              #f
                              diff))
                    ((off x) (if (equal? diff 'off)
                                 #f
                                 diff))
                    (else diff)))))
    (if mask
        (map minus diff mask)
        diff)))

(define (icfp4-build-tag ok id)
  ; interpret icfp4-match-mask's return value -> open/close tag
  (cond ((or (null? id) (void? id) (not id) (not ok))
         null)
        ((pair? id)
         (map (lambda (x) (icfp4-build-tag ok x)) id))
        (else
         (string-append icfp4-tags-start$
                        (if (icfp4-closetag? ok)
                            icfp4-tags-close$
                            "")
                        id
                        icfp4-tags-end$))))



;--- Optimze
(define (icfp4-optimize-redundances open-tags this-block next-block out)
  (cond ((null? this-block)
         null)
        ((null? next-block)
         (list this-block))
        (else 
         (let ((this-pre   (icfp4-outblock-pre  this-block))
               (this-text  (icfp4-outblock-text this-block))
               (this-post  (icfp4-outblock-post this-block))
               (next-pre   (icfp4-outblock-pre  next-block))
               (next-text  (icfp4-outblock-text next-block))
               (next-post  (icfp4-outblock-post next-block)))
           (cond ((or (null? this-post) (null? next-pre))
                  (cons this-block
                        (icfp4-optimize-redundances
                         (icfp4-optimize-opentags
                          open-tags this-pre this-post)
                         next-block
                         (icfp4-list-ref 0 out)
                         (icfp4-nth-dr 0 out))))
                 ((and (pair? this-post) (pair? next-pre)
                       (> (length this-post) 1) (> (length next-pre) 1)
                       (not (equal? this-post next-pre)))
                  (if #t ;(andmap (lambda (x) (member x this-post)) next-pre)
                      (let* ((optimized    (icfp4-optimize-this-redundance 
                                            open-tags this-post 
                                            next-block next-pre out))
                             (this-post-o  (car optimized))
                             (next-pre-o   (cdr optimized))
                             (this-block-o (icfp4-outblock-new this-pre
                                                               this-text
                                                               this-post-o))
                             (next-block-o (icfp4-outblock-new next-pre-o
                                                               next-text
                                                               next-post))
                             (open-tags    (icfp4-optimize-opentags
                                            open-tags this-pre this-post-o)))
                        ; (display "[")(display this-post)(display "=")
                        ; (display this-post-o)(display "]")(newline)
                        (cons 
                         this-block-o 
                         (icfp4-optimize-redundances open-tags next-block-o
                                                     (icfp4-list-ref 0 out)
                                                     (icfp4-nth-dr 0 out))))
                      (cons 
                       this-block (icfp4-optimize-redundances
                                   open-tags  
                                   next-block
                                   (icfp4-list-ref 0 out) 
                                   (icfp4-nth-dr 0 out)))))
                 (else
                  (let* ((optimized    (icfp4-optimize-this-redundance 
                                        open-tags this-post 
                                        next-block next-pre out))
                         (this-post-o  (car optimized))
                         (next-pre-o   (cdr optimized))
                         (this-block-o (icfp4-outblock-new this-pre
                                                           this-text
                                                           this-post-o))
                         (next-block-o (icfp4-outblock-new next-pre-o
                                                           next-text
                                                           next-post))
                         (open-tags    (icfp4-optimize-opentags
                                        open-tags this-pre this-post-o)))
                    (cons this-block-o
                          (icfp4-optimize-redundances 
                           open-tags
                           next-block-o
                           (icfp4-list-ref 0 out)
                           (icfp4-nth-dr 0 out))))))))))

(define (icfp4-optimize-opentags taglst pre post)
  taglst)
(define (icfp4-optimize-opentags-disabled taglst pre post)
  (letrec ((remove-this (lambda (tag lst)
                          (if (null? lst)
                              null
                              (let ((hd (car lst))
                                    (tl (cdr lst)))
                                (if (equal? tag (car hd))
                                    tl
                                    (cons hd (remove-this tag tl)))))))
           (remover (lambda (taglst lst)
                      (if (or (null? taglst) (null? lst))
                          null
                          (remover (cdr taglst)
                                   (remove-this (car taglst) lst))))))
    (let* ((pre    (if (pair? pre) pre (list pre)))
           (post   (if (pair? post) post (list post)))
           (closed (remove* pre post))
           (taglst (remover closed taglst))
           (opened (remove* post pre))
           (taglst (if (null? opened)
                       taglst
                       (append (map (lambda (x) (cons x 0)) opened)
                               taglst)))
           (taglst (if (null? taglst)
                       taglst
                       (mergesort (map (lambda (x)
                                         (cons (car x) (+ (cdr x) 1)))
                                       taglst)
                                  (lambda (a b)
                                    (< (cdr a) (cdr b)))))))
      ; (display "[")(display taglst)
      ; (display ":")(display opened)(display ":")(display closed)
      ; (display ":")(display pre)(display ":")(display post)
      ; (display "]")(newline)
      taglst)))

(define (icfp4-optimize-this-redundance-dbg open-tags this-post 
                                        block next-pre out)
  (display "[")(display this-post)
  (display ":")(display next-pre)
  (let ((rv (icfp4-optimize-this-redundance-1 open-tags this-post 
                                        block next-pre out)))
    (display ":")(display rv)
  (display "]")(newline)
    rv))

(define (icfp4-optimize-this-redundance open-tags this-post 
                                        block next-pre out)
  ;  (display ":")(display this-post)(display "-")(display next-pre)(newline)
  (let* ((chooser (lambda (x) (if (and (pair? x) (= (length x) 1))
                                  (car x)
                                  x)))
         (this (chooser this-post))
         (next (chooser next-pre)))
    (cond ((equal? this next)
           `(,null . ,null))
          ((and (string? this) (string? next)
                (string=? this next))
           `(,null . ,null))
          ((and (string? this) (list? next)
                (member this next))
           `(,null . ,(remove this next)))
          ((and (list? this) (string? next)
                (member next this))
           `(,(remove next this) . ,null))
          ((and (list? this) (list? next))
           (if (andmap (lambda (x) (member x this)) next)
               `(,(remove* next this) . ,null)
               ; <TT>bla</TT></B><B><U>bla</U>
               (let* ((next-next   ; rel-pos of nearest that
                       (let ((ordered (icfp4-optimize-find-posttags
                                       #f (remove* this next) block out)))
                         (icfp4-optimize-find-block-pos (car ordered))))
                      (these (filter (lambda (x) (member x next))
                                     this))
                      (these   ; list of these with ++ range
                       (let* ((ordered (icfp4-optimize-find-posttags
                                        #f these block out))
                              (ordered 
                               (filter
                                (lambda (x)
                                  (<= next-next
                                      (icfp4-optimize-find-block-pos x)))
                                       ordered)))
                         (map (lambda (x) 
                                (icfp4-optimize-find-block-tag x))
                              ordered))))
                 (if (null? these)
                     `(,this . ,next)
                     `(,(remove* these this) 
                       . ,(remove* these next))))))
          (else
           `(,this . ,next)))))

(define (icfp4-optimize-whitespaces-phase2 lst skip?)
  (if (null? lst)
      null
      (let* ((block (car lst))
             (type  (icfp4-block-get-type block)))
        (if (equal? type 'whitespace)
            (let* ((prop (icfp4-block-get-prop block))
                   (tt?  (icfp4-state-tt prop)))
              (if (or tt? (not skip?))
                  (cons block 
                        (icfp4-optimize-whitespaces-phase2 (cdr lst) #t))
                  (icfp4-optimize-whitespaces-phase2 (cdr lst) skip?)))
            (cons block 
                  (icfp4-optimize-whitespaces-phase2 (cdr lst) #f))))))

(define (icfp4-optimize-pl out)
  ; PL shortcut
  ; changing <I><S><TT>foo</TT></S></I> bar <TT><S><I> gee</I></S></TT> 
  ; into     <I><S><TT>foo<PL> bar </PL> gee</TT></S></I> 
  (if (null? out)
      null
      (let* ((block (car out))
             (post  (icfp4-outblock-post block))
             (n     (icfp4-count-member post icfp4-tags-textstyle)))
        (if (>= n 2)
            (let ((opt (icfp4-optimize-pl-optimizer 
                        block
                        (filter (lambda (x) (member x icfp4-tags-textstyle))
                                post)
                        (cdr out))))
              (if opt
                  (let ((opt-block (list-ref opt 0))
                        (opt-lst   (list-ref opt 1))
                        (rest      (list-ref opt 2)))
                    ; (display opt-block)(newline)
                    `(,opt-block ,@opt-lst . ,(icfp4-optimize-pl rest)))
                  (begin
                    (cons block (icfp4-optimize-pl (cdr out))))))
            (cons block (icfp4-optimize-pl (cdr out)))))))

(define (icfp4-optimize-pl-finder tags sub lst)
;  (display "[")(display tags)
;  (display ":")(display sub)
;  (display "]")
  (if (null? lst)
      #f
      (let* ((block (car lst))
             (pre   (icfp4-outblock-pre block))
             (post  (icfp4-outblock-post block)))
; unstable: (let ((found (filter (lambda (x) (member x pre)) tags)))
        (if (andmap (lambda (x) (member x pre)) tags)
            (list tags sub block (cdr lst))
            (if (not (null? post))
                #f
                (icfp4-optimize-pl-finder 
                 tags
                 (if (null? sub) 
                     (list block) `(,@sub ,block))
                        (cdr lst)))))))

(define (icfp4-optimize-pl-optimizer block post lst)
  (let ((found (icfp4-optimize-pl-finder post null lst)))
    (if found
        (let ((subpost  (list-ref found 0))
              (sublst   (list-ref found 1)))
          ; (display "[")(display subpost)
          ; (display ":")(display sublst)
          ; (display ":")(display (list-ref found 2))
          ; (display "]")(newline)
          (if (and (>= (length subpost) 2)
                   (not (null? sublst)))
              (let* ((beg-post (icfp4-outblock-post block))
                     (beg-post (remove* subpost beg-post))
                     (begblock (icfp4-outblock-patch block #f #f beg-post))
                     (endblock (list-ref found 2))
                     (end-pre  (icfp4-outblock-pre endblock))
                     (end-pre  (remove* subpost end-pre))
                     (endblock (icfp4-outblock-patch endblock end-pre #f #f))
                     (pl-begin (car sublst))
                     (pl-pre   (icfp4-outblock-pre pl-begin))
                     (pl-pre   (cons "PL" pl-pre))
                     (pl-begin (icfp4-outblock-patch pl-begin pl-pre #f #f))
                     (sublst   (icfp4-replace-nth sublst 'first pl-begin))
                     (pl-end   (list-ref sublst (- (length sublst) 1)))
                     (pl-post  (icfp4-outblock-post pl-end))
                     (pl-post  (cons "PL" pl-post))
                     (pl-end   (icfp4-outblock-patch pl-end #f #f pl-post))
                     (sublst   (icfp4-replace-nth sublst 'end pl-end))
                     (sublst   `(,@(icfp4-optimize-pl sublst)
                                 ,endblock))
                     (rest     (list-ref found 3)))
                ; (display sublst)(newline)
                `(,begblock 
                  ,sublst
                  ,rest))
              #f))
        #f)))

(define (icfp4-optimize-colour out)
  ; color nesting
  ; changing <r>aaa</r><g>bbb</g><b>ccc</b><g>ddd</g><r>eee</r> 
  ; into     <r>aaa<g>bbb<b>ccc</b>ddd</g>eee</r> 
  (icfp4-optimize-nested icfp4-tags-colour null out))

(define (icfp4-optimize-size out)
  (icfp4-optimize-nested icfp4-tags-size null out))

(define (icfp4-optimize-nested tags open-tags out)
  (if (null? out)
      null
      (let* ((block     (car out))
             (pre       (icfp4-outblock-pre block))
             (nested    (ormap (lambda (x) (member x tags)) pre))
             (drop      (ormap (lambda (x) (member x open-tags)) pre))
             (pre       (if (and nested drop)
                            (remove (car nested) pre)
                            pre))
             (block     (if (and nested drop)
                            (icfp4-outblock-patch block pre #f #f)
                            block))
             (post      (icfp4-outblock-post block))
             (ok        (ormap (lambda (x) (member x tags)) post))
             (rest      (cdr out)))
        (if ok
            (let* ((this (car ok))
                   (ok (icfp4-optimize-nested-check? 
                        tags this null null rest))
                   (ok-to-remove? (list-ref ok 0))
;                   (sub (list-ref ok 1))
;                   (rest (list-ref ok 2))
                   (post (if ok-to-remove?
                             (remove this post)
                             post))
                   (plus (remove* post pre))
                   (minus (remove* pre post))
                   (open-tags (append plus open-tags))
                   (open-tags (remove* minus open-tags)))
;              (display ":")(display ok-to-remove?)(display pre)
;              (display post)(display plus)(display minus)
;              (display open-tags)(newline)
              (if ok-to-remove?
;                  `(,(icfp4-outblock-patch block #f #f (remove this post))
;                    ,@(icfp4-optimize-nested tags open-tags sub)
;                    . ,rest)
                  (cons (icfp4-outblock-patch block #f #f (remove this post))
                        (icfp4-optimize-nested tags open-tags rest))
                  (cons block (icfp4-optimize-nested tags open-tags rest))))
            (cons block (icfp4-optimize-nested tags open-tags rest))))))

(define (icfp4-optimize-nested-check? tags this open-tags sub out)
  (if (null? out)
      (list #f sub out)
      (let* ((block (car out))
             (pre   (icfp4-outblock-pre block)))
        (if (member this pre)
            (list #t sub out)
            (let* ((post   (icfp4-outblock-post block))
                   (diff   (remove* pre post))
                   (opened (ormap (lambda (x) (member x tags)) pre))
                   (closed (ormap (lambda (x) (member x tags)) post)))
              (cond ((ormap (lambda (x) (not (member x open-tags))) 
                            diff)
                     (list #f sub out))
                    ((and (not opened) (null? open-tags))
                     (list #f sub out))
                    (else
                     (let* ((open-tags1 (if opened
                                           (cons (car opened) open-tags)))
                            (open-tags1 (if closed
                                           (remove (car closed) open-tags)))
                            (open-tags (append diff open-tags))
                            (sub       (if (null? sub) 
                                           (list block)
                                           `(,@sub ,block))))
                       (icfp4-optimize-nested-check? 
                        tags this open-tags sub (cdr out))))))))))

(define (icfp4-optimize-checkorder out)
  (if (null? out)
      null
      (let* ((block (car out))
             (pre   (icfp4-outblock-pre  block))
             (rest  (cdr out)))
        (cond ((null? rest)
               (list block))
              ((pair? pre)
               (cons (icfp4-optimize-checkorder-this block pre rest)
                     (icfp4-optimize-checkorder rest)))
              (else
               (cons block (icfp4-optimize-checkorder rest)))))))

(define (icfp4-optimize-checkorder-this block tags out)
  (let* ((order (icfp4-optimize-find-posttags #t tags block out))
         (pre   (map (lambda (x) (icfp4-optimize-find-block-tag x)) order)))
    (icfp4-outblock-new pre
                        (icfp4-outblock-text block)
                        (icfp4-outblock-post block))))

(define (icfp4-optimize-find-posttags mode tags block out)
  (let ((order  (map (lambda (x) (icfp4-optimize-find-posttag 
                                  mode x 0 block out))
                     tags)))
    (if (null? order)
        null
        (mergesort order (lambda (a b)
                           (> (icfp4-optimize-find-block-pos a)
                              (icfp4-optimize-find-block-pos b)))))))
         
(define (icfp4-optimize-find-posttag mode tag count block out)
  (if (null? out)
      (cons count tag)
      (let ((post  (icfp4-outblock-post block)))
        (if (or (equal? tag post)
                (and (pair? post)
                     (let ((val (member tag post)))
                       (if mode
                           val
                           (not val)))))
            (icfp4-optimize-find-block-new count tag)
            (icfp4-optimize-find-posttag
             mode tag (+ count 1) (car out) (cdr out))))))

(define (icfp4-optimize-find-block-new pos tag) (cons pos tag))
(define (icfp4-optimize-find-block-pos block)   (car block))
(define (icfp4-optimize-find-block-tag block)   (cdr block))


;--- Main
(define (icfp4-main exit?)
  (icfp4-parse-cmdline)
  (icfp4-read-init)
  (icfp4-output (icfp4-readall-tokenized))
  (if exit? (exit)))

(define (icfp4-test)
  (with-input-from-file
      "F:/Scheme/tml/ICFP Programming Contest 4/test.smlng"
    (lambda () (icfp4-main #f))))

(define (icfp4-test1)
  (with-input-from-file
      "F:/Scheme/tml/ICFP Programming Contest 4/test-mini.smlng"
    (lambda () (icfp4-main #f))))

(define (icfp4-test2)
  (with-input-from-file
      "F:/Scheme/tml/ICFP Programming Contest 4/test-o.smlng"
    (lambda () (icfp4-main #f))))

(define (icfp4-test3)
  (with-input-from-file
      "F:/Scheme/tml/ICFP Programming Contest 4/test-bug.smlng"
    (lambda () (icfp4-main #f))))

(icfp4-main #t)
; (icfp4-test2)
